@@ -1,70 +1,61 @@
 package Petal::Tiny;
-$Petal::Tiny::VERSION = '1.13';
+$Petal::Tiny::VERSION = '1.16';
 use warnings;
 use strict;
 use Carp;
 
-# REX/Perl 1.0 
+# REX/Perl 1.0
 # Robert D. Cameron "REX: XML Shallow Parsing with Regular Expressions",
-# Technical Report TR 1998-17, School of Computing Science, Simon Fraser 
+# Technical Report TR 1998-17, School of Computing Science, Simon Fraser
 # University, November, 1998.
 # Copyright (c) 1998, Robert D. Cameron.
 # The following code may be freely used and distributed provided that
 # this copyright and citation notice remains intact and that modifications
 # or additions are clearly identified.
-our $TextSE = "[^<]+";
-our $UntilHyphen = "[^-]*-";
-our $Until2Hyphens = "$UntilHyphen(?:[^-]$UntilHyphen)*-";
-our $CommentCE = "$Until2Hyphens>?";
-our $UntilRSBs = "[^\\]]*](?:[^\\]]+])*]+";
-our $CDATA_CE = "$UntilRSBs(?:[^\\]>]$UntilRSBs)*>";
-our $S = "[ \\n\\t\\r]+";
-our $NameStrt = "[A-Za-z_:]|[^\\x00-\\x7F]";
-our $NameChar = "[A-Za-z0-9_:.-]|[^\\x00-\\x7F]";
-our $Name = "(?:$NameStrt)(?:$NameChar)*";
-our $QuoteSE = "\"[^\"]*\"|'[^']*'";
-our $DT_IdentSE = "$S$Name(?:$S(?:$Name|$QuoteSE))*";
-our $MarkupDeclCE = "(?:[^\\]\"'><]+|$QuoteSE)*>";
-our $S1 = "[\\n\\r\\t ]";
-our $UntilQMs = "[^?]*\\?+";
-our $PI_Tail = "\\?>|$S1$UntilQMs(?:[^>?]$UntilQMs)*>";
-our $DT_ItemSE = "<(?:!(?:--$Until2Hyphens>|[^-]$MarkupDeclCE)|\\?$Name(?:$PI_Tail))|%$Name;|$S";
-our $DocTypeCE = "$DT_IdentSE(?:$S)?(?:\\[(?:$DT_ItemSE)*](?:$S)?)?>?";
-our $DeclCE = "--(?:$CommentCE)?|\\[CDATA\\[(?:$CDATA_CE)?|DOCTYPE(?:$DocTypeCE)?";
-our $PI_CE = "$Name(?:$PI_Tail)?";
-our $EndTagCE = "$Name(?:$S)?>?";
-our $AttValSE = "\"[^<\"]*\"|'[^<']*'";
-our $ElemTagCE = "$Name(?:$S$Name(?:$S)?=(?:$S)?(?:$AttValSE))*(?:$S)?/?>?";
-our $ElemTagCE_Mod = "$S($Name)(?:$S)?=(?:$S)?($AttValSE)";
-our $MarkupSPE = "<(?:!(?:$DeclCE)?|\\?(?:$PI_CE)?|/(?:$EndTagCE)?|(?:$ElemTagCE)?)";
-our $XML_SPE = "$TextSE|$MarkupSPE";
+my $TextSE = "[^<]+";
+my $UntilHyphen = "[^-]*-";
+my $Until2Hyphens = "$UntilHyphen(?:[^-]$UntilHyphen)*-";
+my $CommentCE = "$Until2Hyphens>?";
+my $UntilRSBs = "[^\\]]*](?:[^\\]]+])*]+";
+my $CDATA_CE = "$UntilRSBs(?:[^\\]>]$UntilRSBs)*>";
+my $S = "[ \\n\\t\\r]+";
+my $NameStrt = "[A-Za-z_:]|[^\\x00-\\x7F]";
+my $NameChar = "[A-Za-z0-9_:.-]|[^\\x00-\\x7F]";
+my $Name = "(?:$NameStrt)(?:$NameChar)*";
+my $QuoteSE = "\"[^\"]*\"|'[^']*'";
+my $DT_IdentSE = "$S$Name(?:$S(?:$Name|$QuoteSE))*";
+my $MarkupDeclCE = "(?:[^\\]\"'><]+|$QuoteSE)*>";
+my $S1 = "[\\n\\r\\t ]";
+my $UntilQMs = "[^?]*\\?+";
+my $PI_Tail = "\\?>|$S1$UntilQMs(?:[^>?]$UntilQMs)*>";
+my $DT_ItemSE = "<(?:!(?:--$Until2Hyphens>|[^-]$MarkupDeclCE)|\\?$Name(?:$PI_Tail))|%$Name;|$S";
+my $DocTypeCE = "$DT_IdentSE(?:$S)?(?:\\[(?:$DT_ItemSE)*](?:$S)?)?>?";
+my $DeclCE = "--(?:$CommentCE)?|\\[CDATA\\[(?:$CDATA_CE)?|DOCTYPE(?:$DocTypeCE)?";
+my $PI_CE = "$Name(?:$PI_Tail)?";
+my $EndTagCE = "$Name(?:$S)?>?";
+my $AttValSE = "\"[^<\"]*\"|'[^<']*'";
+my $ElemTagCE = "$Name(?:$S$Name(?:$S)?=(?:$S)?(?:$AttValSE))*(?:$S)?/?>?";
+my $ElemTagCE_Mod = "$S($Name)(?:$S)?=(?:$S)?($AttValSE)";
+my $MarkupSPE = "<(?:!(?:$DeclCE)?|\\?(?:$PI_CE)?|/(?:$EndTagCE)?|(?:$ElemTagCE)?)";
+my $XML_SPE = "$TextSE|$MarkupSPE";
 # REX END - thank you Robert for this 26 line XML parser - awesome ...
 
+my $ATTR_RE = qr /$ElemTagCE_Mod/;
 
-our $RE_1 = qr /$ElemTagCE/;
-our $RE_2 = qr /$ElemTagCE_Mod/;
-our $VARIABLE_RE_SIMPLE   = qq |\$[A-Za-z_][A-Za-z0-9_\.:\/]+|;
-our $VARIABLE_RE_BRACKETS = qq |(?<!\$)\\{.*?(?<!\\\\)\\}|;
-our $STRING_TOKEN_RE      = "($VARIABLE_RE_SIMPLE|$VARIABLE_RE_BRACKETS)";
-
-our $TAL = 'petal';
-
-our $STOP_RECURSE = 0;
-
+my $DEFAULT_NS = 'petal';
 
 sub new {
     my $class = shift;
     $class    = ref $class || $class;
     my $thing = shift;
     my $self  = bless {}, $class;
-    if    (defined $thing and $thing =~ /(\<|\n|\>)/) { $self->{xmldata} = $thing }
+    if (defined $thing and $thing =~ /(\<|\n|\>)/) {
+        $self->{xmldata} = $thing;
+    }
     elsif (defined $thing) {
-        $self->{xmldata} = do {
-            open my $xmldatafile, "<", $thing or die "cannot read open $thing";
-            my $xmldata = join '', <$xmldatafile>;
-            close $xmldatafile;
-            $xmldata;
-        };
+        open my $xmldatafile, "<", $thing or die "cannot read open $thing";
+        $self->{xmldata} = join '', <$xmldatafile>;
+        close $xmldatafile;
     }
     return $self;
 }
@@ -75,177 +66,184 @@ sub process {
     my $context = { @_ };
     my $data    = $self->{xmldata};
     defined $data or return; # empty data, empty result.
-    return $self->makeitso($data, $context); # earl grey. hot.
+    return $self->makeitso($self->xml2nodes($data), $context); # earl grey. hot.
 }
 
+sub xml2nodes {
+    my ($self, $xml) = @_;
+
+    my @flat = ( $xml =~ /$XML_SPE/og );
+    my $top = { _kids => [], _ns => $DEFAULT_NS };
+    my @nest = ( $top );
+    for my $tag (@flat) {
+        my $node = tag2node($tag, $nest[-1]{_ns}); # if ns is not explicitly set, inherit parent ns
+
+        if ($node->{_close}) {
+            my $open = pop @nest;
+            confess "Too many close-tags! Last </$node->{_tag}>." if $open == $top;
+            if (lc($node->{_tag}) ne lc($open->{_tag})) {
+                my $in = "";
+                $in .= $nest[$_]{_elem} for 1..$#nest;
+                $in .= $open->{_elem};
+                confess "Wrong close-tag '</$node->{_tag}>' following '$in'";
+            }
+        }
+        else {
+            push @{ $nest[-1]{_kids} }, $node;
+            push @nest, $node unless ($node->{_simple} or $node->{_selfclose});
+        }
+    }
+    confess "Unbalanced tree, more open than close nodes" if @nest > 1;
+
+    my @nodes = @{ $top->{_kids} };
+
+    return \@nodes;
+}
 
 sub makeitso {
-    my $self    = shift;
-    my $xml     = shift;
-    my @xml     = ref $xml ? @{$xml} : ( $xml =~ /$XML_SPE/g );
-    my $context = shift || {};
-    my @head    = ();
-    my @body    = ();
-    my @tail    = ();
-    while (@xml) {
-        my $elem = shift @xml;
-        tag_self_close ($elem) and do {
-            push @body, $elem;
-            @tail = @xml;
-            last;
-        };
-        my $opentag = tag_open ($elem);
-        $opentag and do {
-            push @body, $elem;
-            my $balance = 1;
-            while ($balance) {
-                @xml or confess "cannot find closing tag for $elem";
-                my $elem = shift @xml;
-                tag_open  ($elem) and $balance++;
-                tag_close ($elem) and $balance--;
-                push @body, $elem;
-            }
-            @tail = @xml;
-            last;
-        };
-        tag_close ($elem) and confess "cannot find opening tag for $elem";
+    my ($self, $nodes, $context) = @_;
 
-        $elem = $self->_interpolate_dollar($context, $elem, 'resolve_expression');
-        push @head, $elem;
+    return "" unless @$nodes;
+
+    my @res;
+    for my $node (@$nodes) {
+        if ($node->{_simple}) {
+            push @res, $self->_interpolate_dollar($context, $node->{_elem}, 'resolve_expression');
+        }
+        else {
+            push @res, $self->makeitso_node($node, $context);
+        }
     }
-    my @res = ();
-    push @res, @head                                   if (@head);
-    push @res, $self->makeitso_block(\@body, $context) if (@body);
-    push @res, $self->makeitso(\@tail, $context)       if (@tail);
-    return join '', @res;
-}
 
+    return join "", @res;
+}
 
 sub _interpolate_dollar {
     my ($self, $context, $string, $method) = @_;
 
-    my $subst = sub {
-        my $what = shift;
-        my $res = $self->$method($what, $context);
-        return $res if defined $res;
-        carp "'$what' in \$-interpolation resolved to undef";
-        return "";
-    };
-
     if ($string =~ /\$/) {
-        $string =~ s/(?<!\$) \$\{  ( [^{}]+           ) \}  / $subst->($1) /xegi;
-        $string =~ s/(?<!\$) \$\{? ( [a-z0-9-\/\:\_]+ ) \}? / $subst->($1) /xegi;
+        my $subst = sub {
+            my $what = shift;
+            my $res = $self->$method($what, $context);
+            return $res if defined $res;
+            carp "'$what' in \$-interpolation resolved to undef";
+            return "";
+        };
+
+        $string =~ s/(?<!\$) \$\{  ( [^{}]+         ) \}  / $subst->($1) /xegi;
+        $string =~ s/(?<!\$) \$\{? ( [a-z0-9-\/:_]+ ) \}? / $subst->($1) /xegi;
         $string =~ s/\$\$/\$/g;
     }
     return $string;
 }
 
-
-sub namespace {
-    my $self = shift;
+sub _deep_copy {
     my $node = shift;
-    for my $k (keys %{$node}) {
-        $k =~ /^xmlns\:/ or next;
-        my $v = $node->{$k};
-        if ($v eq 'http://purl.org/petal/1.0/') {
-            delete $node->{$k};
-            $k =~ s/^xmlns\://;
-            return $k;
+    my %copy = %$node;
+    my @kids;
+    for my $kid (@{ $node->{_kids} }) {
+        push @kids, _deep_copy($kid);
+    }
+    $copy{_kids} = \@kids;
+    return \%copy;
+}
+
+sub makeitso_node {
+    my ($self, $node, $context) = @_;
+
+    my $TAL = $node->{_ns};
+
+    my $STOP_RECURSE = 0;
+    
+    if ($node->{_has_tal}) {
+        $node->{_change} = 1;
+
+        if (defined( my $stuff = delete $node->{"$TAL:on-error"} )) {
+            my $nodeCopy = { %$node };
+            my $res = eval { $self->makeitso_node($node, $context); };
+            if ($@) {
+                for my $k (keys %$nodeCopy) { delete $nodeCopy->{$k} if $k =~ /^$TAL:/ }
+                delete $nodeCopy->{_selfclose};
+                $nodeCopy->{_contents} = $self->resolve_expression($stuff, $context);
+                return node2txt($nodeCopy);
+            }
+            return $res;
+        }
+
+        $context = { %$context };
+
+        if (defined( my $stuff = delete $node->{"$TAL:define"} )) {
+            for my $def (split /;(?!;)/, $stuff) {
+                my ($symbol, $expression) = split ' ', $def, 2;
+                $context->{$symbol} = $self->resolve_expression($expression, $context);
+            }
+        }
+
+        if (defined( my $stuff = delete $node->{"$TAL:condition"} )) {
+            for my $cond (split /;(?!;)/, $stuff) {
+                return '' unless $self->resolve_expression($cond, $context);
+            }
+        }
+
+        if (defined( my $stuff = delete $node->{"$TAL:repeat"} )) {
+            my @loops = split /;(?!;)/, $stuff;
+            my $count = 0;
+            return join "", $self->_do_repeat(\$count, 1, \@loops, $node, $context);
+        }
+
+        if (defined( my $stuff = delete $node->{"$TAL:content"} )) {
+            my $res = $self->resolve_expression($stuff, $context);
+            $node->{_contents} = defined $res ? $res : "";
+            delete $node->{_selfclose};
+
+            # set the stop recurse flag so that if content contains $foo and $bar,
+            # those aren't interpolated as variables.
+            $STOP_RECURSE = 1;
+        }
+
+        if (defined( my $stuff = delete $node->{"$TAL:replace"} )) {
+            my $res = $self->resolve_expression($stuff, $context);
+            return defined $res ? $res : '';
+        }
+
+        if (defined( my $stuff = delete $node->{"$TAL:attributes"} )) {
+            for my $att (split /;(?!;)/, $stuff) {
+                my ($symbol, $expression) = split ' ', $att, 2;
+                my $add = ($symbol =~ s/^\+//);
+                my $new = $self->resolve_expression($expression, $context);
+                if (defined $new) {
+                    if ($add) {
+                        my $old = $node->{$symbol};
+                        $old = "" unless defined $old;
+                        $new = $old . $new;
+                    }
+                    $node->{$symbol} = $new;
+                }
+                else {
+                    delete $node->{$symbol} unless $add;
+                }
+            }
+        }
+
+        if (defined(my $stuff = delete $node->{"$TAL:omit-tag"})) {
+            if ($stuff eq '' or $self->resolve_expression($stuff, $context)) {
+                return $node->{_contents} if $STOP_RECURSE;
+                return $self->makeitso($node->{_kids}, $context);
+            }
         }
     }
-    return;
-}
 
-
-sub makeitso_block {
-    my $self    = shift;
-    my $xml     = shift;
-    my $context = shift;
-    my @xml     = ref $xml ? @{$xml} : ( $xml =~ /$XML_SPE/g );
-    my $tag     = shift (@xml);
-    my $gat     = pop (@xml);
-    my $node    = tag_open ($tag) || tag_self_close ($tag);
-    my $ns      = $self->namespace($node);
-    local $TAL  = $ns || $TAL;
-    if (has_instructions ($node)) {
-        $context = { %{$context} };
-        return $self->tal_on_error($node, \@xml, $gat, $context);
+    unless ($STOP_RECURSE) {
+        $node->{_contents} = $self->makeitso($node->{_kids}, $context);
     }
-    else {
-        $tag = node2tag ($node) if ($ns);
-        if ($gat) { return $tag . $self->makeitso(\@xml, $context) . $gat }
-        else      { return $tag                                           } # self-closing tag
-    }
-}
-
-
-sub tal_on_error {
-    my ($self, $node, $xml, $end, $context) = @_;
-    my $stuff   = delete $node->{"$TAL:on-error"};
-    defined $stuff or return $self->tal_define($node, $xml, $end, $context);
-    my $nodeCopy = { %{$node} };
-    my $res = eval { $self->tal_define($node, $xml, $end, $context) };
-    if ($@) {
-        my @result = ();
-        for my $k (keys %{$nodeCopy}) { delete $nodeCopy->{$k} if $k =~ /^$TAL:/ }
-        delete $nodeCopy->{_close} and $end = "</$nodeCopy->{_tag}>"; # deal with self closing tags
-        push @result, node2tag ($nodeCopy);
-        push @result, $self->resolve_expression($stuff, $context);
-        push @result, $end;
-        return join '', @result;
-    }
-    else {
-        return $res;
-    }
-}
-
-
-sub tal_define {
-    my ($self, $node, $xml, $end, $context) = @_;
-    my $stuff   = delete $node->{"$TAL:define"};
-    defined $stuff || return $self->tal_condition($node, $xml, $end, $context);
-    my $newContext = { %{$context} };
-    my $define  = trim ($stuff);
-    for my $def (split /;(?!;)/, $define) {
-        $def = trim($def);
-        my ($symbol, $expression) = split /\s+/, $def, 2;
-        $newContext->{$symbol} = $self->resolve_expression($expression, $newContext);
-    }
-    return $self->tal_condition($node, $xml, $end, $newContext);
-}
-
-
-sub tal_condition {
-    my ($self, $node, $xml, $end, $context) = @_;
-    my $stuff   = delete $node->{"$TAL:condition"};
-    defined $stuff or return $self->tal_repeat($node, $xml, $end, $context);
-
-    my $condition = trim ($stuff);
-    for my $cond (split /;(?!;)/, $condition) {
-        $cond = trim($cond);
-        $self->resolve_expression($cond, $context) or return '';
-    }
-    return $self->tal_repeat($node, $xml, $end, $context);
-}
-
-
-sub tal_repeat {
-    my ($self, $node, $xml, $end, $context) = @_;
-    my $stuff   = delete $node->{"$TAL:repeat"};
-    defined $stuff or return $self->tal_content($node, $xml, $end, $context);
-    
-    my @loops = split /;/, $stuff;
-    my $count = 0;
-    return $self->_do_repeat(\$count, 1, \@loops, $node, $xml, $end, { %$context });
+    return node2txt($node);
 }
 
 sub _do_repeat {
-    my ($self, $count, $last, $loops_ref, $node, $xml, $end, $context) = @_;
-
+    my ($self, $count, $last, $loops_ref, $node, $context) = @_;
     my @loops = @$loops_ref;
     my $stuff = shift @loops;
-    my $repeat = trim ($stuff);
-    my ($symbol, $expression) = split /\s+/, $repeat, 2;
+    my ($symbol, $expression) = split ' ', $stuff, 2;
     my $array  = $self->resolve_expression($expression, $context);
     $array = [ $array ] unless ref $array; # we don't judge
     my @result;
@@ -253,7 +251,7 @@ sub _do_repeat {
         my $item = $array->[$idx];
         $context->{$symbol} = $item;
         if (@loops) {
-            push @result, $self->_do_repeat($count, $last && $idx == $#$array, \@loops, $node, $xml, $end, $context);
+            push @result, $self->_do_repeat($count, $last && $idx == $#$array, \@loops, $node, $context);
         }
         else {
             $$count++;
@@ -265,87 +263,24 @@ sub _do_repeat {
             $context->{repeat}->{start}  = $$count == 1 ? 1 : 0;
             $context->{repeat}->{end}    = $last && $idx == $#$array ? 1 : 0;
             $context->{repeat}->{inner}  = $context->{repeat}->{start} || $context->{repeat}->{end} ? 0 : 1;
-            push @result, $self->tal_content({ %$node }, $xml, $end, $context);
+
+            push @result, $self->makeitso_node(_deep_copy($node), $context);
         }
     }
-    return join '', @result;
-}
-
-
-sub tal_content {
-    my ($self, $node, $xml, $end, $context) = @_;
-    my $stuff   = delete $node->{"$TAL:content"};
-    defined $stuff or return $self->tal_replace($node, $xml, $end, $context);
-    
-    my $res = $self->resolve_expression($stuff, $context);
-    $xml    = defined $res ? [ $res ] : [];
-    delete $node->{_close} and $end = "</$node->{_tag}>"; # deal with self closing tags
-    
-    # set the stop recurse flag so that if content contains $foo and $bar,
-    # those aren't interpolated as variables.
-    local ( $STOP_RECURSE ) = ( 1 );
-    return $self->tal_replace($node, $xml, $end, $context);
-}
-
-
-sub tal_replace {
-    my ($self, $node, $xml, $end, $context) = @_;
-    my $stuff   = delete $node->{"$TAL:replace"};
-    defined $stuff or return $self->tal_attributes($node, $xml, $end, $context);
-    my $res = $self->resolve_expression($stuff, $context);
-    return defined $res ? $res : '';
-}
-
-
-sub tal_attributes {
-    my ($self, $node, $xml, $end, $context) = @_;
-    my $stuff = delete $node->{"$TAL:attributes"};
-    defined $stuff or return $self->tal_omit_tag($node, $xml, $end, $context);
-   
-    my $attributes  = trim ($stuff);
-    for my $att (split /;(?!;)/, $attributes) {
-        $att = trim ($att);
-        my ($symbol, $expression) = split /\s+/, $att, 2;
-        my $add = ($symbol =~ s/^\+//);
-        my $new = $self->resolve_expression($expression, $context);
-        if (defined $new) {
-            if ($add) {
-                my $old = $node->{$symbol};
-                $old = "" unless defined $old;
-                $new = $old . $new;
-            }
-            $node->{$symbol} = $new;
-        }
-        else {
-            delete $node->{$symbol} unless $add;
-        }
-    }
-    return $self->tal_omit_tag($node, $xml, $end, $context);
-}
-
-
-sub tal_omit_tag {
-    my ($self, $node, $xml, $end, $context) = @_;
-    my $stuff  = delete $node->{"$TAL:omit-tag"};
-    my $omit   = defined $stuff ? do { $stuff eq '' ? 1 : $self->resolve_expression($stuff, $context) } : undef;
-    $omit and not $end and return ''; # omit-tag on a self-closing tag means *poof*, nothing left
-    my @result = ();
-    push @result, node2tag ($node) unless ($omit);
-    if ($end) {
-        push @result, do { $STOP_RECURSE ? join '', @{$xml} : $self->makeitso($xml, $context) };
-        push @result, $end unless ($omit);
-    }
-    return join '', @result;
+    return @result;
 }
 
 
 sub resolve_expression {
-    my $self    = shift;
-    my $expr    = trim(shift);
-    my $context = shift || confess "resolve_expression() : no context";
-    $expr =~ s/\;\;/;/g;
-    $expr =~ s/\$\$/\$/g;
-    $expr eq 'nothing' and return undef;    
+    my ($self, $expr, $context) = @_;
+
+    $expr = "" unless defined $expr;
+    $expr =~ s/[\n\r]/ /g;
+    $expr =~ s/^\s+//;
+    $expr =~ s/\s+$//;
+
+    $expr =~ s/([;\$])\1/$1/g;
+    $expr eq 'nothing' and return undef;
     $expr =~ s/^fresh\s+//;
     my $structure = ($expr =~ s/^structure\s+//);
     my $resolved = $self->resolve($expr, $context);
@@ -358,25 +293,20 @@ sub reftype {
 }
 
 sub resolve {
-    my $self    = shift;
-    my $expr    = trim(shift);
-    my $context = shift || confess "resolve() : no context";
-    $expr =~ /:(?!pattern)/ and do {
-        my ($mod, $expr) = split /:(?!pattern)/, $expr, 2;
-        my $meth = "modifier_$mod";
-        $self->can("modifier_$mod") and return $self->$meth($expr, $context);
+    my ($self, $expr, $context) = @_;
+    $expr =~ /:(?!pattern)/ and do { # XXX what is :pattern?
+        my ($mod, $expr) = split /:(?!pattern)\s*/, $expr, 2;
+        my $meth = $self->can("modifier_$mod");
+        return $self->$meth($expr, $context) if $meth;
         confess "unknown modifier $mod";
     };
-    $expr =~ /^--/ and do {
-      $expr =~ s/^--//;
-      return $expr;
-    };
-    $expr =~ s/\r/ /g;
-    my ($what, @args) = split /\s+/, $expr;
+    return $expr if $expr =~ s/^--//;
+
+    my ($what, @args) = split ' ', $expr;
     defined $what or return;
-    
+
     my (@path)   = split /\//, $what;
-    my @resolved = ();
+    my @resolved;
     my $obj      = $context;
     @args        = map { $self->resolve($_, $context) } @args;
     while (@path) {
@@ -436,9 +366,9 @@ sub resolve {
 
 
 sub modifier_true {
-    my $self = shift;
-    my $arg  = $self->resolve(shift(), shift());
-    ref $arg and $self->reftype($arg) eq 'ARRAY' and return @{$arg};
+    my ($self, $expr, $context) = @_;
+    my $arg  = $self->resolve($expr, $context);
+    ref $arg and $self->reftype($arg) eq 'ARRAY' and return scalar @$arg;
     return $arg ? 1 : 0;
 }
 
@@ -450,122 +380,112 @@ sub modifier_false {
 
 
 sub modifier_string {
-    my $self    = shift;
-    my $string  = shift;
-    my $context = shift;
+    my ($self, $string, $context) = @_;
     $string = $self->_interpolate_dollar($context, $string, 'resolve');
     return $string;
 }
 
 
-sub node2tag {
+sub node2txt {
     my $node  = shift;
-    my $tag   = delete $node->{_tag};
-    for (keys %{$node}) { /^($TAL:)/ and delete $node->{$_} }
-    my $open  = delete $node->{_open}  || 0;
-    my $close = delete $node->{_close} || 0;
-    my $att   = join ' ', map { qq|$_="$node->{$_}"| } keys %{$node};
-    $open  and $close and return $att ? "<$tag $att />" : "<$tag />";
-    $close and return "</$tag>";
-    $open  and return $att ? "<$tag $att>" : "<$tag>";
-    die "There is probably a bug somewhere. A tag that's not and open tag and not a close tag?";
+
+    return $node unless ref $node eq 'HASH'; # handle textnodes introduced in makeitso_node
+    return $node->{_elem} if $node->{_simple};
+
+    delete $node->{_ns};
+    delete $node->{_has_tal};
+    delete $node->{_kids};
+
+    my $change   = delete $node->{_change};
+    my $elem     = delete $node->{_elem};
+    my $tag      = delete $node->{_tag};
+    my $close    = delete $node->{_selfclose};
+    my $quotes   = delete $node->{_quotes};
+    my $contents = delete $node->{_contents};
+    my $att      = join ' ', map { my $q = $quotes->{$_} || '"'; qq|$_=$q$node->{$_}$q| } keys %$node;
+
+    if ($close) {
+        return $change ? ($att ? "<$tag $att />" : "<$tag />") : $elem;
+    }
+
+    my $start = $change ? ($att ? "<$tag $att>" : "<$tag>") : $elem;
+    my $end   = "</$tag>";
+
+    $contents = "" unless defined $contents;
+
+    return $start . $contents . $end;
 }
 
+sub tag2node {
+    my ($elem, $ns) = @_;
 
-sub trim {
-    my $string = shift;
-    $string or return $string;
-    $string =~ s/\r//g;
-    $string =~ s/\n/ /g;
-    $string =~ s/^\s+//;
-    $string =~ s/\s+$//;
-    return $string;
-}
+    if (my ($has_close,     $tag) = $elem =~ m, \A < ( /? ) ( [A-Za-z0-9] [A-Za-z0-9_:-]* ) ,x and
+        my ($has_self_close     ) = $elem =~ m, ( /? ) > \z ,x)
+    {
+        return { _tag => $tag, _close => 1 } if $has_close; # don't waste any time on </...> nodes, they're just for book-keeping
 
+        my %node          = extract_attributes($elem);
+        $node{_ns}      ||= $ns;
 
-sub has_instructions {
-    my $node = shift;
-    return grep /^$TAL:/, keys %{$node};
-}
+        $node{_has_tal}   = exists $node{_ns_prefix}{ $node{_ns} };
+        $node{_tag}       = $tag;
+        $node{_selfclose} = $has_self_close;
+        $node{_elem}      = $elem;
+        $node{_kids}      = [];
 
+        delete $node{_ns_prefix};
 
-sub tag_open {
-    my $elem = shift;
-    my $node = undef;
-    not defined $elem and confess ('undefined elem');
+        return \%node;
+    }
 
-    $elem !~ /^<\!/ and
-    $elem !~ /^<\// and
-    $elem !~ /\/>$/ and
-    $elem !~ /^<\?/ and
-    $elem =~ /^</   and do {
-	my %node      = extract_attributes ($elem);
-	($node{_tag}) = $elem =~ /.*?([A-Za-z0-9][A-Za-z0-9_:-]*)/;
-	$node{_open}  = 1;
-	$node{_close} = 0;
-	$node = \%node;
+    return {
+        _elem => $elem,
+        _simple => 1,
     };
-    return $node;
 }
 
-
-sub tag_close {
-    my $elem = shift;
-    my $node = undef;
-    $elem !~ /^<\!/ and
-    $elem =~ /^<\// and
-    $elem !~ /\/>$/ and do {
-        my %node      = ();
-	($node{_tag}) = $elem =~ /.*?([A-Za-z0-9][A-Za-z0-9_:-]*)/;
-	$node{_open}  = 0;
-	$node{_close} = 1;
-	$node = \%node;
-    };
-    return $node;
-}
-
-
-sub tag_self_close {
-    my $elem = shift;
-    my $node = undef;
-    $elem !~ /^<\!/ and
-    $elem !~ /^<\// and
-    $elem =~ /\/>$/ and
-    $elem =~ /^</   and do {
-	my %node      = extract_attributes ($elem);
-        ($node{_tag}) = $elem =~ /.*?([A-Za-z0-9][A-Za-z0-9_:-]*)/;
-	$node{_open}  = 1;
-	$node{_close} = 1;
-	$node = \%node;
-    };
-    return $node;
-}
-
-
 sub extract_attributes {
     my $tag = shift;
-    my ($tags) = $tag =~ /$RE_1/g;
-    my %attr = $tag =~ /$RE_2/g;
-    foreach my $key (keys %attr)
-    {
-        my $val = $attr{$key};
-        $val    =~ s/^(\"|\')//;
-        $val    =~ s/(\"|\')$//;
-        $attr{$key} = $val;
+
+    my %attr = $tag =~ /$ATTR_RE/og;
+
+    my (%quotes, %prefix);
+
+    foreach my $key (keys %attr) {
+        $attr{$key} =~ s/^(['"])(.*?)\1$/$2/;
+        my $q = $1 || '"';
+
+        if ($key =~ /^(.*?):/) {
+            if ($1 eq 'xmlns' && $attr{$key} eq 'http://purl.org/petal/1.0/') {
+                delete $attr{$key};
+                $key           =~ s/^xmlns\://;
+                $attr{_ns}     = $key;
+                $attr{_change} = 1;
+                next;
+            }
+            $prefix{$1} = 1;
+        }
+        $quotes{$key} = $q;
     }
-    
+
+    $attr{_quotes} = \%quotes;
+    $attr{_ns_prefix} = \%prefix;
+
     %attr;
 }
 
+my %_encode_dict = (
+    '&' => '&amp;',
+    '<' => '&lt;',
+    '>' => '&gt;',
+    '"' => '&quot;',
+    "'" => '&apos;',
+);
 
 sub xmlencode {
     my $string = shift;
     return $string if !$string or ref $string;
-    $string =~ s/&/&amp;/g;
-    $string =~ s/</&lt;/g;
-    $string =~ s/>/&gt;/g;
-    $string =~ s/"/&quot;/g;
-    $string =~ s/'/&apos;/g;
+    $string =~ s/([&<>"'])/$_encode_dict{$1}/g;
     return $string;
 }
 
@@ -750,7 +670,7 @@ things such as:
     <p>Checkout amount: <span petal:content="self/basket/total">TOTAL</span> USD</p>
 
 You can still write:
-    
+
     <!-- BAM! Petal way. Much easier, especially for quick prototyping. -->
     <p>Checkout amount: $self/basket/total USD</p>
 
@@ -767,7 +687,7 @@ tal:attributes always overrides the content of an attribute, but
 occasionally you want to concatenate the new string to the existing
 string. Prefixing the attribute name with '+' allows you do to this:
 
- <div class="foo " tal:attribute="+class bar"/>
+ <div class="foo " tal:attributes="+class bar"/>
 
 outputs
 
@@ -811,7 +731,7 @@ In the following examples, we'll assume that the template is used as follows:
   print $template->process ( $hashref );
 
 Then we will show how the Petal Expression Syntax maps to the Perl way of
-accessing these values.  
+accessing these values.
 
 
 =head2 accessing scalar values
@@ -1005,7 +925,7 @@ and returns the value.
 Alternatively, you could write:
 
   string:Welcome ${user/real_name}, it is ${date}!
-  
+
 The advantage of using curly brackets is that it lets you interpolate
 expressions which invoke methods with parameters, i.e.
 
